  (* +--------------------------------------------------------------------+
     |                                                                    |
     |  Parseur pour un langage  oprateurs prfixe, infixe et postfixe  |
     |                                                                    |
     |  M. Quercia, mars 1996                                             |
     +--------------------------------------------------------------------+ *)


(*
 * [expr p flux] value le dbut de [flux] jusqu' rencontrer un
 *   oprateur infixe de priorit infrieure ou gale  [p].
 *
 * [atome flux] envoie la premire valeur dans [flux] en appliquant
 *   les oprateurs prfixe et les regroupement entre parenthses.
 *
 * [liste flux] et [fin_liste flux] assemblent une liste de valeurs.
 *
 * [suite_expr p a flux] applique les oprateurs postfixe et les oprateurs
 *   infixe de priorit suprieure  [p]  la valeur [a].
 *
 * [infixe_prioritaire p flux] renvoie un'oprateur infixe s'il figure
 *   en tte de [flux] et a une priorit suprieure  [p], choue sinin.
 *)

let rec expr p = function [< atome a; (suite_expr p a) b >] -> b

and atome = function
| [< 'Valeur v >] -> v
| [< 'Prfixe f; atome v >] -> f v
| [< 'Lpar; (expr 0) v; 'Rpar >] -> v
| [< 'Liste f; 'Lpar; liste l; 'Rpar >] -> f l

and liste = function
| [< (expr 0) v; fin_liste l >] -> v :: l
| [< >] -> []

and fin_liste = function
| [< 'Lsep; (expr 0) v; fin_liste l >] -> v :: l
| [< >] -> []

and suite_expr p a = function
| [< 'Postfixe f; (suite_expr p (f a)) b >] -> b
| [< (infixe_prioritaire p) (q,f); (expr q) b; (suite_expr p (f a b)) c >] -> c
| [< >] -> a

and infixe_prioritaire p flux = match stream_get flux with
| Infixe(q,f),_ when q > p -> stream_next flux; (q,f)
| _ -> raise Parse_failure
;; 

(*
 * [analyse f lexer s] analyse la chaine [s] et renvoie la valeur qu'elle
 * reprsente. [s] est d'abord convertie en flux de caractres, ce flux est
 * soumis  l'analyseur lexical [lexer], puis les lexmes obtenus sont
 * convertis en lments syntaxiques par [f].
 *)

let analyse syntax_of_lexme lexer s =
  let i = ref (-1) in
  let ch() = incr i; if !i < string_length s then s.[!i] else raise Parse_failure in
  let st = lexer(stream_from ch) in
  let stx() = syntax_of_lexme(stream_next st) in
  try match stream_from stx with [< (expr 0) v; end_of_stream _ >] -> v
  with ex -> raise(Erreur(ex,!i))
;;
